#include "fuzzygroup.h"
#include "fuzzyrule.h"
#include <stdarg.h>

#define DFZ_CENTROID 0
#define DFZ_BISECTOR 1
#define DFZ_LOM 2
#define DFZ_MOM 3
#define DFZ_HOM 4

#define OPS_ZADEH 0
#define OPS_LARSEN 1
#define OPS_LUKASIEWITZ 2

/** Trieda reprezentuje fuzzy regulator.
*	Umonuuje na stavit vstupne a vystupne mnoziny. Poskytuje metody na ziskanie vystupu.
*
*/

class cFuzzyObj
{
	private:

	public:
		int inputs;		///< uchovava pocet vstupnych fuzzygroup (lingvistickych premennych )
		int outputs;	///< uchovava pocet vystupnych fuzzygroup (lingvistickych premennych)
		int rules;		///< uchovava pocet pridanych pravidiel
		int samples;	///< urcuje pocet samplov na intervale

		fuzzygroup **input;		///< reprezentuje pole ukazovatelov na fuzzygroup-y
		fuzzygroup **output;	///< reprezentuje pole ukazovatelov na fuzzygroup-y
		fuzzyrule **rule;		///< reprezentuje pole ukazovatelov na pravydla fuzzyrule
	

		/** Konstruktor fuzzyregulatora.
		*	nastavy hodnoty stavovych premennych na 0.
		*/
		cFuzzyObj();
		~cFuzzyObj();
		
		/** Nastavy pocet vzoriek, ktore maju byt na intervale.
		*
		*	@param samples1 pocet vzoriek
		*/
		void SetSamples(int samples1);

		/** Zadefinuje mnozinu <b>fuzzygroup</b> pre vstupne fuzzy mnoziny na rovnakom intervale.
		*
		*	@param min1 spodna hranica intervalu
		*	@param max1 horna hranica intervalu
		*/
		void AddInput(float min1,float max1);

		/** Zadefinuje mnozinu <b>fuzzygroup</b> pre vystupne fuzzy mnoziny na rovnakom intervale.
		*
		*	@param min1 spodna hranica intervalu
		*	@param max1 horna hranica intervalu
		*/
		void AddOutput(float min1,float max1);

		/** Prida fuzzy mnozinu do zvolenej vstupnej fuzzygroup.
		*	
		*	@param inputgroup1 je to index fuzzygroupy, do ktorej sa ma pridat
		*			prislusna fuzzy mnozina. Tento index suhlasi s poradim pridavania
		*			fuzzygroup do regulatora. Indexovat sa zacina od 0.
		*	@param ... ďalsim parametrom je typ charakteristickej funkcie pre fuzzy mnozinu
		*	@param ... ďalsimi parametrami su hodony na intervale <min1, max1> pre vytvorenie
		*			konkretnej charakteristickej funkcie.
		*
		*	@see Konstanty pre typy charakteristickych funkcii v subore fuzzy.h
		*/
		void AddInputMF(int inputgroup1,...);

		/** Prida fuzzy mnozinu do zvolenej vystupnej fuzzygroup.
		*	
		*	@param outputgroup1 je to index fuzzygroupy, do ktorej sa ma pridat
		*			prislusna fuzzy mnozina. Tento index suhlasi s poradim pridavania
		*			fuzzygroup do regulatora. Indexovat sa zacina od 0.
		*	@param ... ďalsim parametrom je typ charakteristickej funkcie pre fuzzy mnozinu
		*	@param ... ďalsimi parametrami su hodony na intervale <min1, max1> pre vytvorenie
		*			konkretnej charakteristickej funkcie.
		*
		*	@see Konstanty pre typy charakteristickych funkcii v subore fuzzy.h
		*/
		void AddOutputMF(int outputgroup1,...);

		/** Prida pravidlo do regulatora.
		*
		*	Okrem parametra <b>outputgroup1</b> je potrebne zadat aj ďalsie parametre.
		*	Ich pocet je urceny:
		*		- poctom vstupnych fuzzygroup (lingvistickych premennych)
		*		- plus jeden vystupny parameter pre vystpnu fuzzygroup-u (lingvisticku premennu)
		*		.
		*	Pre kazdu vstupnu fuzzygroup-u je potrebne zadat index charakteristickej funkcie,
		*	ktora sa bude pri vyhodnocovani pravidla uplatnovat. Rovnako to plati pre charakteristicku
		*	funkciu vystupnej fuzzygroupy.
		*
		*	@param outputgroup1 urcuje index vystupnej fuzzygroup-y.
		*/
		void AddRule(int outputgroup1,...);


		/** Vyhodnoti vystup z regulatora na zaklade vlozenych vstupov.
		*
		*	Okrem parametrov <b>outputgroup1</b>, <b>method</b>, <b>ops</b>, su potrebne aj parametre
		*	vstupnych hodnot pre fuzzy regulator. Ich pocet je zhodny s poctom vstupnych fuzzygroup
		*	(lingvistickych premennych).
		*
		*	@param outputgroup1 fuzzygroup-a (lingvisticka premenna), ktora nas zaujima. Napriklad sila.
		*	@param method metoda defuzifikacie
		*	@param ops typ aplikovanej logiky
		*	@param ... ostatne parametre
		*
		*	@return vystup z regulatora
		*/
		float CountResult(int outputgroup1,int method,int ops,...);
}; 
